
--------------------------------------------------------------------------
----------------- STRING package compilation file ------------------------
---------------------- name is stringpack.ads ----------------------------
--------------------------------------------------------------------------

--------------------------------------------------------------------------
------------------  package specification StringPack ---------------------
--------------------------------------------------------------------------

WITH Text_IO;

PACKAGE StringPack IS

-- simple set of operations for handling strings of bounded, but
-- varying lengths. Operations include concatenation, conversion
-- from and to decimal representation of numbers, and extraction of
-- substrings. As a general rule for all the operations, strings
-- are truncated and sizes of other parameters are reduced so that
-- the result conforms to the types in question. Therefore, the
-- package does not contain exception handlers for the predefined
-- exceptions. If neither truncation nor size reduction is possible
-- the exception StringPack_error is raised. The generic paramter
-- maxvarlg is the upper bound for the length of the varstrings.

StringPack_Error: EXCEPTION;


maxvarlg : CONSTANT := 100;

SUBTYPE varstr IS STRING(1..maxvarlg);
TYPE VARSTRING IS RECORD
                    str : varstr;
                    lg  : INTEGER := 0;
                  END RECORD;

-- a VARSTRING variable v represents the STRING v.str(1..v.lg).
-- make sure that 0 <= lg <= maxvarlg holds true!

varnull : CONSTANT VARSTRING := ((OTHERS => '$'),0);


-------------------- basic STRING operations ------------------
---------------------------------------------------------------

FUNCTION  "&" (va: VARSTRING; vb: VARSTRING) RETURN VARSTRING;
FUNCTION  "&" (va: STRING;    vb: VARSTRING) RETURN VARSTRING;
FUNCTION  "&" (va: VARSTRING; vb: STRING)    RETURN VARSTRING;

-- convert INTEGER to minimal STRING representation and concatenate
FUNCTION "&" (va: VARSTRING; i:INTEGER) RETURN VARSTRING;
FUNCTION "&" (i:INTEGER; va: VARSTRING) RETURN VARSTRING;
FUNCTION "&" (s:STRING; i:INTEGER)      RETURN VARSTRING;
FUNCTION "&" (i:INTEGER; s:STRING)      RETURN VARSTRING;

-- convert FLOAT to minimal STRING representation and concatenate
FUNCTION "&" (va: VARSTRING; f:FLOAT) RETURN VARSTRING;
FUNCTION "&" (f:FLOAT; va: VARSTRING) RETURN VARSTRING;
FUNCTION "&" (s:STRING; f:FLOAT)      RETURN VARSTRING;
FUNCTION "&" (f:FLOAT; s:STRING)      RETURN VARSTRING;

-- equality for varstrings
FUNCTION equal (v: VARSTRING; s: STRING)      RETURN BOOLEAN;
FUNCTION equal (s: STRING; v: VARSTRING)      RETURN BOOLEAN;
FUNCTION equal (va: VARSTRING; vb: VARSTRING) RETURN BOOLEAN;
-- note that predefined "=" does not yield the desired result.

-- conversion functions STRING <--> VARSTRING
FUNCTION  "+" (v: VARSTRING) RETURN STRING;
FUNCTION  "+"  (v: STRING)   RETURN VARSTRING;

-- repeat character given number of times and RETURN result
FUNCTION "*" (n: INTEGER; c: CHARACTER) RETURN VARSTRING;

-- slice of a VARSTRING
FUNCTION slice (v: VARSTRING; from, to: INTEGER) RETURN VARSTRING;

----------------------- conversion functions ------------------------
---------------------------------------------------------------------

-- convert INTEGER to VARSTRING
FUNCTION  cvis (item  : INTEGER;
    width: INTEGER := 0; filler: CHARACTER := ' '; base: POSITIVE := 10)
     RETURN VARSTRING;
-- computes minimum length STRING to represent the given
-- INTEGER using the given base. RETURNs this STRING if
-- width > maxvarlg or if width is smaller than the length of
-- this STRING.  otherwise, pads this STRING on the left
-- with filler characters to obtain the given width and RETURNs
-- the resulting STRING. note that the sliding of "-" depends on
-- the filler character, so that both equal("-005",cvis(5,4,'0'))
-- and equal("  -5",cvis(5,4,' ')) yield "true". Embedded "_" are
-- allowed.

-- convert FLOAT to VARSTRING
FUNCTION  cvfs (item: FLOAT; fore: INTEGER:= 0; aft: INTEGER := 2)
      RETURN VARSTRING;
-- computes minimum length STRING to represent the given FLOATing point
-- number without exponent and given size of fraction. RETURNs this
-- STRING if fore does not exceed the number of digits before the
-- decimal point. otherwise, pads this STRING on the left with blank
-- characters to obtain the given width and RETURNs the resulting
-- STRING. the conversion algorithm is based on computing
-- item * 10**aft and rounding this number to a long_INTEGER number.
-- If aft is too large for this method, then a smaller aft is used
-- automatically. StringPack_error is raised if item is outside the
-- range of long_INTEGER.

-- convert LONG_INTEGER to VARSTRING
FUNCTION  cvls (item  : LONG_INTEGER;
                width : INTEGER := 0;
                filler: CHARACTER := ' ';
		base  : POSITIVE := 10) RETURN VARSTRING;

-- convert BOOLEAN to STRING
cvbs : constant array(BOOLEAN) of VARSTRING :=
   (false => ((1 => '0', others => '$'), 1),
    true  => ((1 => '1', others => '$'), 1));

-- convert VARSTRING to INTEGER
FUNCTION cvsi (item: VARSTRING) RETURN INTEGER;
-- the given STRING must be the decimal representation
-- of an INTEGER possibly including a sign character
-- the INTEGER may be surrounded by blanks. raises StringPack_
-- error in case of syntax violation.

-- convert VARSTRING to FLOAT
FUNCTION cvsf(item: IN VARSTRING) RETURN FLOAT;
-- the given STRING must be the decimal representation of a
-- FLOAT number. sign, fractional part and exponent parts are optional.
-- embedded "_" are allowed.

-- convert character to BOOLEAN
cvcb: constant array(CHARACTER) of BOOLEAN :=
  (CHARACTER'FIRST .. '0'| '2'..CHARACTER'LAST => false, '1' => true);

-- convert cases of alphabetic characters, l=lower, u=upper
FUNCTION cvlu (item: VARSTRING) RETURN VARSTRING;   -- lower -> upper
FUNCTION cvul (item: VARSTRING) RETURN VARSTRING;   -- upper -> lower


----------------------  date and time -------------------------
---------------------------------------------------------------

FUNCTION timestamp RETURN VARSTRING;
-- RETURNs the current date and time as a STRING of length 20 in the
-- format >>yy-mm-dd hh:mm:ss.ss<< in decimal notation with
-- leading zeroes


-----------------  read and Print procedures -----------------
---------------------------------------------------------------

-- the following procedures provide a simple means of input
-- and output. names are chosen so that they do not interfere with
-- text_io.

inbuffer, outbuffer : VARSTRING;

-- Print v on a new line
PROCEDURE Print(v: VARSTRING := outbuffer);
PROCEDURE Print(s: STRING) renames text_io.put_line;
PROCEDURE Write(v: VARSTRING := outbuffer);
PROCEDURE Write(s: STRING) RENAMES Text_IO.Put;

-- Print empty lines
PROCEDURE NewLine(n: INTEGER := 1);

-- advance to new page
PROCEDURE newpage renames text_io.new_page;

-- read next line into inbuffer
PROCEDURE read;

-- check for END of file
FUNCTION endoffile RETURN BOOLEAN renames text_io.end_of_file;
enderror: exception renames text_io.end_error;



--------------------- input from varstrings -------------------------
---------------------------------------------------------------------

-- the following three functions delete a prefix of length width from
-- inbuffer and convert it to the requested result type. if 0 <=
-- inbuffer.lg < width then the entire inbuffer is taken as the
-- prefix. the prefix is empty if width <= 0 and StringPack_error
-- is raised for GetInt and GetFloat.

FUNCTION GetText (width: INTEGER) RETURN VARSTRING;
FUNCTION GetInt (width: INTEGER) RETURN INTEGER;
FUNCTION GetFloat (width: INTEGER) RETURN FLOAT;


-- the following three functions start by reading lines (possibly none)
-- via read into inbuffer until inbuffer contains a nonblank character.
-- then they search for the first nonblank sequence of characters in
-- inbuffer. this sequence is converted to the requested type and
-- RETURNed as result. as a side effect the prefix of inbuffer up to
-- and including this sequence is deleted from inbuffer.

FUNCTION GetText RETURN VARSTRING;
FUNCTION GetInt RETURN INTEGER;
FUNCTION GetFloat RETURN FLOAT;

FUNCTION GetChar RETURN CHARACTER;
FUNCTION GetBool RETURN BOOLEAN;
-- same as above with "nonblank sequence of characters" replaced by
-- "nonblank character". GetBool yields "true" for a '1' and "false"
-- for any other nonblank character.

-- checks whether line is all blanks
FUNCTION checkblanks (line: VARSTRING := inbuffer) RETURN BOOLEAN;

PROCEDURE locatetext
     (left,right: out INTEGER; line: VARSTRING:= inbuffer);
-- locate the next nonblank sequence of characters and RETURN its
-- bounds in left and right. line is left unchanged. If the line is
-- all blanks then values such that right < left are RETURNed.

END StringPack;
