-- Realisierung der Tupel-Funktionen

with Ada.Tags; use Ada.Tags;
with stringpack; use stringpack;

package body Tuple is
    -- First: liefert ersten Parameter
    -- Uebergabe: Tupel
    -- Rueckgabe: Iterator mit Tupel-Header und Zeiger auf 1. Parameter
    function First (T: Tuple_Type) return Iterator_Type is
    begin
        return (T.Header, T.Header.First);
    end First;

    -- Succ:  liefert naechsten Parameter
    -- Uebergabe: Iterator mit Tupel-Header und Zeiger auf aktuellen Parameter
    -- Rueckgabe: Iterator mit Tupel-Header und Zeiger auf naechsten Parameter
    function Succ (I: Iterator_Type) return Iterator_Type is
    begin
        return (I.Header, I.Current.Next);
    end Succ;

    -- Value: liefert Wert des aktuellen Parameters
    -- Uebergabe: Iterator mit Tupel-Header und Zeiger auf aktuellen Parameter
    -- Rueckgabe: Inhalt des aktuellen Parameters
    function Value (I: Iterator_Type) return Parameter_Type'Class is
    begin
        return I.Current.Para.all;
    end Value;

    -- End_Of_Tupel: zeigt an, ob der letzte Parameter erreicht ist
    -- Uebergabe: Iterator mit Tupel-Header und Zeiger auf aktuellen Parameter
    -- Rueckgabe: Boolean: true oder false
    function End_Of_Tuple (I: Iterator_Type) return boolean is
    begin
        return I.Current = null;
    end;

    -- Insert_First: fuegt vorne an die Parameter-Liste einen an
    -- Uebergabe: Tupel (= Parameter-Liste) und den neuen Parameter
    procedure Insert_First (T: Tuple_Type; P: Parameter_Type'Class) is
    begin
        T.Header.First := new Parameter_Record'
                              (new Parameter_Type'Class'(P), T.Header.First);
        if   T.Header.Last = null
        then T.Header.Last := T.Header.First;
        end if;
    end Insert_First;

    procedure Insert_Last (T: Tuple_Type; P: Parameter_Type'Class) is
    begin
        if    T.Header.Last = null
        then  T.Header.First := new Parameter_Record'
                              (new Parameter_Type'Class'(P), null);
              T.Header.Last := T.Header.First;
        else  T.Header.Last.Next := new Parameter_Record'
                                    (new Parameter_Type'Class'(P), null);
              T.Header.Last := T.Header.Last.Next;
        end if;
    end Insert_Last;

    -- Delete_First: loescht den ersten Parameter der Parameter-Liste
    -- Uebergabe: das Tupel (= die Parameter-Liste)
    procedure Delete_First (T: Tuple_Type) is
    begin
        T.Header.First := T.Header.First.Next;
        if   T.Header.First = null
        then T.Header.Last := null;
        end if;
    end Delete_First;

    procedure Empty (T: in out Tuple_Type) is
    begin
        T.Header.First := null;
        T.Header.Last := null;
    end Empty;

    -- Match: ueberprueft, ob 2 Tupel uebereinstimmen
    -- Uebergabe: 2 Tupel
    -- Rueckgabe: Boolean, True oder False
    function Match (A, B: Tuple_Type) return boolean is
        I1, I2: Iterator_Type;
    begin
        I1 := First (A);     -- Start von A
        I2 := First (B);     -- Start von B

        -- beide Tupel durchgehen und Parameter fuer Parameter vergleichen
        while not End_Of_Tuple (I1) and not End_Of_Tuple (I2)
        loop
            -- wenn ein Parameter formal ist, darf der andere nicht formal sein
            if I1.Current.Para.Formal and I2.Current.Para.Formal then
                -- print ("Beide formal!");
                return false;
            end if;

            -- die Typen der Parameter muessen uebereinstimmen
            if   I1.Current.Para'tag /= I2.Current.Para'tag then
               -- print ("Tags passen nicht!");
               -- print ("Para1: " & Expanded_Name (I1.Current.Para'tag));
               -- print ("Para2: " & Expanded_Name (I2.Current.Para'tag));
               return false;
            end if;

            if     not I1.Current.Para.Formal
               and not I2.Current.Para.Formal
               and not (I1.Current.Para.all = I2.Current.Para.all) then
               -- print ("Inhalte der Parameter passen nicht!" & Expanded_Name (I1.Current.Para'tag));
               return false;
            end if;

            I1 := Succ (I1);
            I2 := Succ (I2);
            -- print ("Ein Parameter passt!");
        end loop;

        if   not End_Of_Tuple (I1) or not End_Of_Tuple (I2)
        then return false;
             -- print ("Ein Tupel ist laenger!");
        end if;
        -- print ("Tupel-Match!");
        return true;
    end;

    -- Bind: wie Match, zusaetzlich werden bei Uebereinstimmung die
    --       formalen Parameter an aktuelle gebunden
    -- Uebergabe: 2 Tupel
    -- Rueckgabe: ein neues Tupel mit den gebundenen Parametern
    function Bind (A, B: Tuple_Type) return Tuple_Type is
        I1, I2: Iterator_Type;
        C: Tuple_Type;
    begin
        -- print ("... vorm Match ...");
        if not Match (A, B) then      -- wenns nicht zomm-passen, retour
           raise Tuple_Does_Not_Match;
        end if;

        I1 := First (A);     -- Start von A
        I2 := First (B);     -- Start von B

        -- print ("... jetzt gehn ma's durch und machen C ...");
        -- beide Tupel durchgehen und Parameter fuer Parameter binden
        while not End_Of_Tuple (I1) and not End_Of_Tuple (I2)
        loop
           -- Parameter A ist formal --> Inhalt von B einfuegen
           if I1.Current.Para.Formal then
              Insert_Last(C, I2.Current.Para.all);
           -- Sonst A einfuegen
           else
              Insert_Last (C, I1.Current.Para.all);
           end if;
           I1 := Succ (I1);
           I2 := Succ (I2);
        end loop;

        -- print ("vor return in Bind");
        return C;
    end;

    procedure Adjust (T: in out Tuple_Type) is
        PA: Record_Access;
    begin
        -- print ("Adjusting");
        T.Header := new Tuple_Header'(T.Header.all);
        if   T.Header.First /= null
        then T.Header.First := new Parameter_Record'(T.Header.First.all);
             PA := T.Header.First;
             while PA.all.Next /= null
             loop  PA.all.Next := new Parameter_Record'(PA.all.Next.all);
                   PA := PA.Next;
             end loop;
             T.Header.Last := PA;
        end if;
    end Adjust;

end Tuple;
